Index: apc_main.c
===================================================================
--- apc_main.c	(revision 298155)
+++ apc_main.c	(working copy)
@@ -41,6 +41,7 @@
 #include "apc_stack.h"
 #include "apc_zend.h"
 #include "apc_pool.h"
+#include "apc_string.h"
 #include "SAPI.h"
 #include "php_scandir.h"
 #include "ext/standard/php_var.h"
@@ -189,6 +190,7 @@
          */
         status = zend_lookup_class_ex(cl.parent_name,
                                     strlen(cl.parent_name), 
+                                    NULL,
                                     0,
                                     &parent_ptr TSRMLS_CC);
         if (status == FAILURE) {
@@ -396,6 +398,30 @@
 /* }}} */
 
 /* {{{ apc_compile_cache_entry  */
+static zend_op_array *apc_call_old_compile_file(zend_file_handle* h, int type TSRMLS_DC)
+{
+#ifdef ZEND_ENGINE_2_4
+    zend_op_array *op_array;
+    int rebailout;
+   
+    apc_interned_strings_disable(TSRMLS_C);
+    zend_try {
+        op_array = old_compile_file(h, type TSRMLS_CC);
+        rebailout = 0;
+    } zend_catch {
+        rebailout = 1;
+    } zend_end_try();
+    apc_interned_strings_enable(TSRMLS_C);
+    if (rebailout) {
+        zend_bailout();
+    }
+    return op_array;
+#else
+    return old_compile_file(h, type TSRMLS_CC); 
+#endif
+}
+
+/* {{{ apc_compile_cache_entry  */
 zend_bool apc_compile_cache_entry(apc_cache_key_t key, zend_file_handle* h, int type, time_t t, zend_op_array** op_array, apc_cache_entry_t** cache_entry TSRMLS_DC) {
     int num_functions, num_classes;
     apc_function_t* alloc_functions;
@@ -411,7 +437,7 @@
     /* compile the file using the default compile function,  *
      * we set *op_array here so we return opcodes during     *
      * a failure.  We should not return prior to this line.  */
-    *op_array = old_compile_file(h, type TSRMLS_CC);
+    *op_array = apc_call_old_compile_file(h, type TSRMLS_CC);
     if (*op_array == NULL) {
         return FAILURE;
     }
@@ -498,7 +524,7 @@
     int bailout=0;
 
     if (!APCG(enabled) || apc_cache_busy(apc_cache)) {
-        return old_compile_file(h, type TSRMLS_CC);
+        return apc_call_old_compile_file(h, type TSRMLS_CC);
     }
 
     /* check our regular expression filters */
@@ -506,10 +532,10 @@
         int ret = apc_regex_match_array(APCG(compiled_filters), h->opened_path);
 
         if(ret == APC_NEGATIVE_MATCH || (ret != APC_POSITIVE_MATCH && !APCG(cache_by_default))) {
-            return old_compile_file(h, type TSRMLS_CC);
+            return apc_call_old_compile_file(h, type TSRMLS_CC);
         }
     } else if(!APCG(cache_by_default)) {
-        return old_compile_file(h, type TSRMLS_CC);
+        return apc_call_old_compile_file(h, type TSRMLS_CC);
     }
     APCG(current_cache) = apc_cache;
 
@@ -522,7 +548,7 @@
 
     /* try to create a cache key; if we fail, give up on caching */
     if (!apc_cache_make_file_key(&key, h->filename, PG(include_path), t TSRMLS_CC)) {
-        return old_compile_file(h, type TSRMLS_CC);
+        return apc_call_old_compile_file(h, type TSRMLS_CC);
     }
 
 
@@ -542,7 +568,7 @@
                                                 apc_sma_protect, apc_sma_unprotect);
         if (!ctxt.pool) {
             apc_wprint("Unable to allocate memory for pool.");
-            return old_compile_file(h, type TSRMLS_CC);
+            return apc_call_old_compile_file(h, type TSRMLS_CC);
         }
         ctxt.copy = APC_COPY_OUT_OPCODE;
 
@@ -584,14 +610,14 @@
 #ifdef __DEBUG_APC__
                 fprintf(stderr,"Stat failed %s - bailing (%s) (%d)\n",h->filename,SG(request_info).path_translated);
 #endif
-                return old_compile_file(h, type TSRMLS_CC);
+                return apc_call_old_compile_file(h, type TSRMLS_CC);
             }
         }
         if (APCG(max_file_size) < fileinfo.st_buf.sb.st_size) { 
 #ifdef __DEBUG_APC__
             fprintf(stderr,"File is too big %s (%ld) - bailing\n", h->filename, fileinfo.st_buf.sb.st_size);
 #endif
-            return old_compile_file(h, type TSRMLS_CC);
+            return apc_call_old_compile_file(h, type TSRMLS_CC);
         }
         key.mtime = fileinfo.st_buf.sb.st_mtime;
     }
@@ -602,7 +628,7 @@
     if(APCG(write_lock)) {
         if(!apc_cache_write_lock(apc_cache)) {
             HANDLE_UNBLOCK_INTERRUPTIONS();
-            return old_compile_file(h, type TSRMLS_CC);
+            return apc_call_old_compile_file(h, type TSRMLS_CC);
         }
     }
 #endif
@@ -792,6 +818,10 @@
     }
 #endif
 
+#ifdef ZEND_ENGINE_2_4
+    apc_interned_strings_init(TSRMLS_C);
+#endif
+
     APCG(initialized) = 1;
     return 0;
 }
@@ -838,6 +868,10 @@
     apc_cache_destroy(apc_user_cache);
     apc_sma_cleanup();
 
+#ifdef ZEND_ENGINE_2_4
+    apc_interned_strings_shutdown(TSRMLS_C);
+#endif
+
     APCG(initialized) = 0;
     return 0;
 }
Index: php_apc.c
===================================================================
--- php_apc.c	(revision 298155)
+++ php_apc.c	(working copy)
@@ -237,6 +237,9 @@
 STD_PHP_INI_BOOLEAN("apc.enabled",      "1",    PHP_INI_SYSTEM, OnUpdateBool,              enabled,         zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.shm_segments",   "1",    PHP_INI_SYSTEM, OnUpdateShmSegments,       shm_segments,    zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.shm_size",       "32M",  PHP_INI_SYSTEM, OnUpdateShmSize,           shm_size,        zend_apc_globals, apc_globals)
+#ifdef ZEND_ENGINE_2_4
+STD_PHP_INI_ENTRY("apc.shm_strings_buffer", "4M",   PHP_INI_SYSTEM, OnUpdateInt,           shm_strings_buffer,        zend_apc_globals, apc_globals)
+#endif
 STD_PHP_INI_BOOLEAN("apc.include_once_override", "0", PHP_INI_SYSTEM, OnUpdateBool,     include_once,    zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.num_files_hint", "1000", PHP_INI_SYSTEM, OnUpdateInt,            num_files_hint,  zend_apc_globals, apc_globals)
 STD_PHP_INI_ENTRY("apc.user_entries_hint", "4096", PHP_INI_SYSTEM, OnUpdateInt,          user_entries_hint, zend_apc_globals, apc_globals)
Index: apc_compile.c
===================================================================
--- apc_compile.c	(revision 298155)
+++ apc_compile.c	(working copy)
@@ -33,6 +33,7 @@
 #include "apc_compile.h"
 #include "apc_globals.h"
 #include "apc_zend.h"
+#include "apc_string.h"
 #include "ext/standard/php_var.h"
 #include "ext/standard/php_smart_str.h"
 
@@ -236,6 +237,21 @@
 }
 /* }}} */
 
+#ifdef ZEND_ENGINE_2_4
+static char *apc_string_pmemcpy(char *str, size_t len, apc_pool* pool)
+{
+    if (pool->type != APC_UNPOOL) {
+        char * ret = apc_new_interned_string(str, len TSRMLS_CC);
+        if (ret) {
+            return ret;
+        }
+    }
+    return apc_pmemcpy(str, len, pool);
+}
+#else
+# define apc_string_pmemcpy(str, len, pool) apc_pmemcpy(str, len, pool)
+#endif
+
 /* {{{ my_copy_zval */
 static zval* my_copy_zval(zval* dst, const zval* src, apc_context_t* ctxt)
 {
@@ -283,7 +299,7 @@
     case IS_CONSTANT:
     case IS_STRING:
         if (src->value.str.val) {
-            CHECK(dst->value.str.val = apc_pmemcpy(src->value.str.val,
+            CHECK(dst->value.str.val = apc_string_pmemcpy(src->value.str.val,
                                                    src->value.str.len+1,
                                                    pool));
         }
@@ -318,7 +334,35 @@
 }
 /* }}} */
 
+#ifdef ZEND_ENGINE_2_4
 /* {{{ my_copy_znode */
+static void my_check_znode(zend_uchar op_type, apc_context_t* ctxt)
+{
+    assert(op_type == IS_CONST ||
+           op_type == IS_VAR ||
+           op_type == IS_CV ||
+           op_type == IS_TMP_VAR ||
+           op_type == IS_UNUSED);
+}
+/* }}} */
+
+/* {{{ my_copy_zend_op */
+static zend_op* my_copy_zend_op(zend_op* dst, zend_op* src, apc_context_t* ctxt)
+{
+    assert(dst != NULL);
+    assert(src != NULL);
+
+    memcpy(dst, src, sizeof(src[0]));
+
+    my_check_znode(dst->result_type & ~EXT_TYPE_UNUSED, ctxt);
+    my_check_znode(dst->op1_type, ctxt);
+    my_check_znode(dst->op2_type, ctxt);
+
+    return dst;
+}
+/* }}} */
+#else
+/* {{{ my_copy_znode */
 static znode* my_copy_znode(znode* dst, znode* src, apc_context_t* ctxt)
 {
     assert(dst != NULL);
@@ -364,6 +408,7 @@
     return dst;
 }
 /* }}} */
+#endif
 
 /* {{{ my_copy_function */
 static zend_function* my_copy_function(zend_function* dst, zend_function* src, apc_context_t* ctxt)
@@ -468,7 +513,7 @@
          * string of the form:
          *      \0<classname>\0<membername>\0
          */
-        CHECK((dst->name = apc_pmemcpy(src->name, src->name_length+1, pool)));
+        CHECK((dst->name = apc_string_pmemcpy(src->name, src->name_length+1, pool)));
     }
 
 #if defined(ZEND_ENGINE_2) && PHP_MINOR_VERSION > 0
@@ -536,11 +581,11 @@
     dst->class_name = NULL;
 
     if (src->name) {
-        CHECK((dst->name = apc_pmemcpy(src->name, src->name_len+1, pool)));
+        CHECK((dst->name = apc_string_pmemcpy(src->name, src->name_len+1, pool)));
     }
 
     if (src->class_name) {
-        CHECK((dst->class_name = apc_pmemcpy(src->class_name, src->class_name_len+1, pool)));
+        CHECK((dst->class_name = apc_string_pmemcpy(src->class_name, src->class_name_len+1, pool)));
     }
 
     return dst;
@@ -762,9 +807,32 @@
         }
 
         /* create a copy of the bucket 'curr' */
+#ifdef ZEND_ENGINE_2_4
+        if (!curr->nKeyLength) {
+            CHECK((newp = (Bucket*) apc_pmemcpy(curr, sizeof(Bucket), pool)));
+        } else if (IS_INTERNED(curr->arKey)) {
+            CHECK((newp = (Bucket*) apc_pmemcpy(curr, sizeof(Bucket), pool)));
+        } else if (pool->type != APC_UNPOOL) {
+            char *arKey;
+            TSRMLS_FETCH();
+
+            arKey = apc_new_interned_string(curr->arKey, curr->nKeyLength TSRMLS_CC);
+            if (!arKey) {
+                CHECK((newp = (Bucket*) apc_pmemcpy(curr, sizeof(Bucket) + curr->nKeyLength, pool)));
+                newp->arKey = ((char*)newp) + sizeof(Bucket);
+            } else {
+                CHECK((newp = (Bucket*) apc_pmemcpy(curr, sizeof(Bucket), pool)));
+                newp->arKey = arKey;
+            }
+        } else {
+            CHECK((newp = (Bucket*) apc_pmemcpy(curr, sizeof(Bucket) + curr->nKeyLength, pool)));
+            newp->arKey = ((char*)newp) + sizeof(Bucket);
+        }        
+#else
         CHECK((newp = (Bucket*) apc_pmemcpy(curr,
                                   sizeof(Bucket) + curr->nKeyLength - 1,
                                   pool)));
+#endif
 
         /* insert 'newp' into the linked list at its hashed index */
         if (dst->arBuckets[n]) {
@@ -858,7 +926,11 @@
         switch (zo->opcode) {
             case ZEND_JMP:
                 /*Note: if src->opcodes != dst->opcodes then we need to the opline according to src*/
+#ifdef ZEND_ENGINE_2_4
+                zo->op1.jmp_addr = dst->opcodes + (zo->op1.jmp_addr - src->opcodes);
+#else
                 zo->op1.u.jmp_addr = dst->opcodes + (zo->op1.u.jmp_addr - src->opcodes);
+#endif
                 break;
             case ZEND_JMPZ:
             case ZEND_JMPNZ:
@@ -867,7 +939,11 @@
 #ifdef ZEND_ENGINE_2_3
             case ZEND_JMP_SET:
 #endif
+#ifdef ZEND_ENGINE_2_4
+                zo->op2.jmp_addr = dst->opcodes + (zo->op2.jmp_addr - src->opcodes);
+#else
                 zo->op2.u.jmp_addr = dst->opcodes + (zo->op2.u.jmp_addr - src->opcodes);
+#endif
                 break;
             default:
                 break;
@@ -931,6 +1007,22 @@
                                       sizeof(src->refcount[0]),
                                       pool));
 
+#ifdef ZEND_ENGINE_2_4
+    if (src->literals) {
+        zend_literal *p, *q, *end;
+
+        q = src->literals;
+        p = dst->literals = (zend_literal*) apc_pool_alloc(pool, sizeof(zend_literal) * src->last_literal);
+        end = p + src->last_literal;
+        while (p < end) {
+            *p = *q;
+            my_copy_zval(&p->constant, &q->constant, ctxt);
+            p++;
+            q++;
+        }
+    }
+#endif
+
     /* deep-copy the opcodes */
     CHECK(dst->opcodes = (zend_op*) apc_pool_alloc(pool, sizeof(zend_op) * src->last));
 
@@ -966,14 +1058,24 @@
                 if(PG(auto_globals_jit) && flags != NULL)
                 {
                      /* The fetch is only required if auto_globals_jit=1  */
+#ifdef ZEND_ENGINE_2_4
+                    if(zo->extended_value & ZEND_FETCH_TYPE_MASK == ZEND_FETCH_GLOBAL &&
+                            zo->op1_type == IS_CONST && 
+                            Z_TYPE_P(zo->op1.zv) == IS_STRING) {
+                        if (Z_STRVAL_P(zo->op1.zv)[0] == '_') {
+# define SET_IF_AUTOGLOBAL(member) \
+    if(!strcmp(Z_STRVAL_P(zo->op1.zv), #member)) \
+        flags->member = 1 /* no ';' here */
+#else
                     if(zo->op2.u.EA.type == ZEND_FETCH_GLOBAL &&
                             zo->op1.op_type == IS_CONST && 
                             zo->op1.u.constant.type == IS_STRING) {
                         znode * varname = &zo->op1;
                         if (varname->u.constant.value.str.val[0] == '_') {
-#define SET_IF_AUTOGLOBAL(member) \
+# define SET_IF_AUTOGLOBAL(member) \
     if(!strcmp(varname->u.constant.value.str.val, #member)) \
         flags->member = 1 /* no ';' here */
+#endif
                             SET_IF_AUTOGLOBAL(_GET);
                             else SET_IF_AUTOGLOBAL(_POST);
                             else SET_IF_AUTOGLOBAL(_COOKIE);
@@ -981,10 +1083,17 @@
                             else SET_IF_AUTOGLOBAL(_ENV);
                             else SET_IF_AUTOGLOBAL(_FILES);
                             else SET_IF_AUTOGLOBAL(_REQUEST);
+#ifdef ZEND_ENGINE_2_4
                             else if(zend_is_auto_global(
+                                            Z_STRVAL_P(zo->op1.zv),
+                                            Z_STRLEN_P(zo->op1.zv)
+                                            TSRMLS_CC))
+#else
+                            else if(zend_is_auto_global(
                                             varname->u.constant.value.str.val,
                                             varname->u.constant.value.str.len
                                             TSRMLS_CC))
+#endif
                             {
                                 flags->unknown_global = 1;
                             }
@@ -993,18 +1102,30 @@
                 }
                 break;
             case ZEND_RECV_INIT:
+#ifdef ZEND_ENGINE_2_4
+                if(zo->op2_type == IS_CONST &&
+                    Z_TYPE_P(zo->op2.zv) == IS_CONSTANT_ARRAY) {
+#else
                 if(zo->op2.op_type == IS_CONST &&
                     zo->op2.u.constant.type == IS_CONSTANT_ARRAY) {
+#endif
                     if(flags != NULL) {
                         flags->deep_copy = 1;
                     }
                 }
                 break;
             default:
+#ifdef ZEND_ENGINE_2_4
+                if((zo->op1_type == IS_CONST &&
+                    Z_TYPE_P(zo->op1.zv) == IS_CONSTANT_ARRAY) ||
+                    (zo->op2_type == IS_CONST &&
+                        Z_TYPE_P(zo->op2.zv) == IS_CONSTANT_ARRAY)) {
+#else
                 if((zo->op1.op_type == IS_CONST &&
                     zo->op1.u.constant.type == IS_CONSTANT_ARRAY) ||
                     (zo->op2.op_type == IS_CONST &&
                         zo->op2.u.constant.type == IS_CONSTANT_ARRAY)) {
+#endif
                     if(flags != NULL) {
                         flags->deep_copy = 1;
                     }
@@ -1016,18 +1137,47 @@
             return NULL;
         }
 
-/* This code breaks apc's rule#1 - cache what you compile */
+#ifdef ZEND_ENGINE_2_4
+        if (zo->op1_type == IS_CONST) {
+            dst->opcodes[i].op1.literal = src->opcodes[i].op1.literal - src->literals + dst->literals;
+        }
+        if (zo->op2_type == IS_CONST) {
+            dst->opcodes[i].op2.literal = src->opcodes[i].op2.literal - src->literals + dst->literals;
+        }
+        if (zo->result_type == IS_CONST) {
+            dst->opcodes[i].result.literal = src->opcodes[i].result.literal - src->literals + dst->literals;
+        }
+#endif
+
+        /* This code breaks apc's rule#1 - cache what you compile */
         if((APCG(fpstat)==0) && APCG(canonicalize)) {
+#ifdef ZEND_ENGINE_2_4
             if((zo->opcode == ZEND_INCLUDE_OR_EVAL) && 
+                (zo->op1_type == IS_CONST && Z_TYPE_P(zo->op1.zv) == IS_STRING)) {
+                /* constant includes */
+                if(!IS_ABSOLUTE_PATH(Z_STRVAL_P(zo->op1.zv),Z_STRLEN_P(&zo->op1.zv))) { 
+                    if (apc_search_paths(Z_STRVAL_P(zo->op1.zv), PG(include_path), &fileinfo) == 0) {
+#else
+            if((zo->opcode == ZEND_INCLUDE_OR_EVAL) && 
                 (zo->op1.op_type == IS_CONST && zo->op1.u.constant.type == IS_STRING)) {
                 /* constant includes */
                 if(!IS_ABSOLUTE_PATH(Z_STRVAL_P(&zo->op1.u.constant),Z_STRLEN_P(&zo->op1.u.constant))) { 
                     if (apc_search_paths(Z_STRVAL_P(&zo->op1.u.constant), PG(include_path), &fileinfo) == 0) {
+#endif
                         if((fullpath = realpath(fileinfo.fullpath, canon_path))) {
                             /* everything has to go through a realpath() */
                             zend_op *dzo = &(dst->opcodes[i]);
+#ifdef ZEND_ENGINE_2_4
+                            dzo->op1.literal = (zend_literal*) apc_pool_alloc(pool, sizeof(zend_literal));
+                            Z_STRLEN_P(dzo->op1.zv) = strlen(fullpath);
+                            Z_STRVAL_P(dzo->op1.zv) = apc_pstrdup(fullpath, pool);
+                            Z_SET_REFCOUNT_P(dzo->op1.zv, 2);
+                            Z_SET_ISREF(dzo->op1.zv);
+                            dzo->op1.literal->hash_value = zend_hash_func(Z_STRVAL_P(dzo->op1.zv), Z_STRLEN_P(dzo->op1.zv)+1);
+#else
                             dzo->op1.u.constant.value.str.len = strlen(fullpath);
                             dzo->op1.u.constant.value.str.val = apc_pstrdup(fullpath, pool);
+#endif
                         }
                     }
                 }
@@ -1066,7 +1216,7 @@
         for(i = 0; i <  src->last_var; i++) dst->vars[i].name = NULL;
 
         for(i = 0; i <  src->last_var; i++) {
-            CHECK(dst->vars[i].name = apc_pmemcpy(src->vars[i].name,
+            CHECK(dst->vars[i].name = apc_string_pmemcpy(src->vars[i].name,
                                 src->vars[i].name_len + 1,
                                 pool));
         }
@@ -1209,8 +1359,20 @@
 /* }}} */
 
 /* Used only by my_prepare_op_array_for_execution */
-#define APC_PREPARE_FETCH_GLOBAL_FOR_EXECUTION()                                                \
+#ifdef ZEND_ENGINE_2_4
+# define APC_PREPARE_FETCH_GLOBAL_FOR_EXECUTION()                                               \
                          /* The fetch is only required if auto_globals_jit=1  */                \
+                        if(zo->extended_value & ZEND_FETCH_TYPE_MASK == ZEND_FETCH_GLOBAL &&    \
+                            zo->op1_type == IS_CONST &&                                         \
+                            Z_TYPE_P(zo->op1.zv) == IS_STRING &&                                \
+                            Z_STRVAL_P(zo->op1.zv)[0] == '_') {                                 \
+                            (void)zend_is_auto_global(Z_STRVAL_P(zo->op1.zv),                   \
+                                                      Z_STRLEN_P(zo->op1.zv)                    \
+                                                      TSRMLS_CC);                               \
+                        }
+#else
+# define APC_PREPARE_FETCH_GLOBAL_FOR_EXECUTION()                                               \
+                         /* The fetch is only required if auto_globals_jit=1  */                \
                         if(zo->op2.u.EA.type == ZEND_FETCH_GLOBAL &&                            \
                             zo->op1.op_type == IS_CONST &&                                      \
                             zo->op1.u.constant.type == IS_STRING &&                             \
@@ -1220,7 +1382,8 @@
                             (void)zend_is_auto_global(varname->u.constant.value.str.val,        \
                                                           varname->u.constant.value.str.len     \
                                                           TSRMLS_CC);                           \
-                        }                                                                       \
+                        }
+#endif
 
 /* {{{ my_prepare_op_array_for_execution */
 static int my_prepare_op_array_for_execution(zend_op_array* dst, zend_op_array* src, apc_context_t* ctxt TSRMLS_DC) 
@@ -1257,6 +1420,25 @@
 
     if(needcopy) {
 
+#ifdef ZEND_ENGINE_2_4
+        if (src->literals) {
+            zend_literal *p, *q, *end;
+
+            q = src->literals;
+            p = dst->literals = (zend_literal*) apc_xmemcpy(src->literals,
+                                        sizeof(zend_literal) * src->last_literal,
+                                        apc_php_malloc);
+            end = p + src->last_literal;
+            while (p < end) {
+                if (Z_TYPE(q->constant) == IS_CONSTANT_ARRAY) {
+                    my_copy_zval(&p->constant, &q->constant, ctxt);
+                }
+                p++;
+                q++;
+            }
+        }
+#endif
+
         dst->opcodes = (zend_op*) apc_xmemcpy(src->opcodes,
                                     sizeof(zend_op) * src->last,
                                     apc_php_malloc);
@@ -1264,6 +1446,17 @@
         dzo = dst->opcodes;
         while(i > 0) {
 
+#ifdef ZEND_ENGINE_2_4
+            if(zo->op1_type == IS_CONST) {
+               dzo->op1.literal = zo->op1.literal - src->literals + dst->literals;
+            }
+            if(zo->op2_type == IS_CONST) {
+               dzo->op2.literal = zo->op2.literal - src->literals + dst->literals;
+            }
+            if(zo->result_type == IS_CONST) {
+               dzo->result.literal = zo->result.literal - src->literals + dst->literals;
+            }
+#else
             if( ((zo->op1.op_type == IS_CONST &&
                   zo->op1.u.constant.type == IS_CONSTANT_ARRAY)) ||
                 ((zo->op2.op_type == IS_CONST &&
@@ -1273,11 +1466,17 @@
                     assert(0); /* emalloc failed or a bad constant array */
                 }
             }
+#endif
 
             switch(zo->opcode) {
                 case ZEND_JMP:
+#ifdef ZEND_ENGINE_2_4
+                    dzo->op1.jmp_addr = dst->opcodes +
+                                            (zo->op1.jmp_addr - src->opcodes);
+#else
                     dzo->op1.u.jmp_addr = dst->opcodes +
                                             (zo->op1.u.jmp_addr - src->opcodes);
+#endif
                     break;
                 case ZEND_JMPZ:
                 case ZEND_JMPNZ:
@@ -1286,8 +1485,13 @@
 #ifdef ZEND_ENGINE_2_3
                 case ZEND_JMP_SET:
 #endif
+#ifdef ZEND_ENGINE_2_4
+                    dzo->op2.jmp_addr = dst->opcodes +
+                                            (zo->op2.jmp_addr - src->opcodes);
+#else
                     dzo->op2.u.jmp_addr = dst->opcodes +
                                             (zo->op2.u.jmp_addr - src->opcodes);
+#endif
                     break;
                 case ZEND_FETCH_R:
                 case ZEND_FETCH_W:
Index: apc_php.h
===================================================================
--- apc_php.h	(revision 298155)
+++ apc_php.h	(working copy)
@@ -47,6 +47,9 @@
 #include "zend_hash.h"
 #include "zend_extensions.h"
 
+#if ZEND_MODULE_API_NO >= 20100409
+#define ZEND_ENGINE_2_4
+#endif
 #if ZEND_MODULE_API_NO > 20060613
 #define ZEND_ENGINE_2_3
 #endif
Index: apc_zend.c
===================================================================
--- apc_zend.c	(revision 298155)
+++ apc_zend.c	(working copy)
@@ -50,6 +50,38 @@
 
 #define APC_EX_T(offset)                    (*(temp_variable *)((char*)execute_data->Ts + offset))
 
+#ifdef ZEND_ENGINE_2_4
+static zval *apc_get_zval_ptr(zend_uchar op_type, znode_op *node, zval **freeval, zend_execute_data *execute_data TSRMLS_DC)
+{
+    *freeval = NULL;
+
+    switch (op_type) {
+        case IS_CONST:
+            return node->zv;
+        case IS_VAR:
+            return APC_EX_T(node->var).var.ptr;
+        case IS_TMP_VAR:
+            return (*freeval = &APC_EX_T(node->var).tmp_var);
+        case IS_CV:
+        {
+            zval ***ret = &execute_data->CVs[node->var];
+
+            if (!*ret) {
+                zend_compiled_variable *cv = &EG(active_op_array)->vars[node->var];
+
+                if (zend_hash_quick_find(EG(active_symbol_table), cv->name, cv->name_len+1, cv->hash_value, (void**)ret)==FAILURE) {
+                    apc_nprint("Undefined variable: %s", cv->name);
+                    return &EG(uninitialized_zval);
+                }
+            }
+            return **ret;
+        }
+        case IS_UNUSED:
+        default:
+            return NULL;
+    }
+}
+#else
 static zval *apc_get_zval_ptr(znode *node, zval **freeval, zend_execute_data *execute_data TSRMLS_DC)
 {
     *freeval = NULL;
@@ -82,6 +114,7 @@
             return NULL;
     }
 }
+#endif
 
 static int ZEND_FASTCALL apc_op_ZEND_INCLUDE_OR_EVAL(ZEND_OPCODE_HANDLER_ARGS)
 {
@@ -94,12 +127,19 @@
     int ret = 0;
     apc_opflags_t* flags = NULL;
 
+#ifdef ZEND_ENGINE_2_4
+    if (opline->extended_value != ZEND_INCLUDE_ONCE &&
+        opline->extended_value != ZEND_REQUIRE_ONCE) {
+        return apc_original_opcode_handlers[APC_OPCODE_HANDLER_DECODE(opline)](ZEND_OPCODE_HANDLER_ARGS_PASSTHRU);
+    }
+#else
     if (Z_LVAL(opline->op2.u.constant) != ZEND_INCLUDE_ONCE &&
         Z_LVAL(opline->op2.u.constant) != ZEND_REQUIRE_ONCE) {
         return apc_original_opcode_handlers[APC_OPCODE_HANDLER_DECODE(opline)](ZEND_OPCODE_HANDLER_ARGS_PASSTHRU);
     }
+#endif
 
-    inc_filename = apc_get_zval_ptr(&opline->op1, &freeop1, execute_data TSRMLS_CC);
+    inc_filename = apc_get_zval_ptr(opline->op1_type, &opline->op1, &freeop1, execute_data TSRMLS_CC);
     if (Z_TYPE_P(inc_filename) != IS_STRING) {
         tmp_inc_filename = *inc_filename;
         zval_copy_ctor(&tmp_inc_filename);
@@ -120,10 +160,17 @@
     }
 
     if (zend_hash_exists(&EG(included_files), realpath, strlen(realpath) + 1)) {
+#ifdef ZEND_ENGINE_2_4
+        if (!(opline->result_type & EXT_TYPE_UNUSED)) {
+            ALLOC_INIT_ZVAL(APC_EX_T(opline->result.var).var.ptr);
+            ZVAL_TRUE(APC_EX_T(opline->result.var).var.ptr);
+        }
+#else
         if (!(opline->result.u.EA.type & EXT_TYPE_UNUSED)) {
             ALLOC_INIT_ZVAL(APC_EX_T(opline->result.u.var).var.ptr);
             ZVAL_TRUE(APC_EX_T(opline->result.u.var).var.ptr);
         }
+#endif
         if (inc_filename == &tmp_inc_filename) {
             zval_dtor(&tmp_inc_filename);
         }
@@ -147,9 +194,15 @@
         /* Since the op array is a local copy, we can cheat our way through the file inclusion by temporarily 
          * changing the op to a plain require/include, calling its handler and finally restoring the opcode.
          */
+#ifdef ZEND_ENGINE_2_4
+        opline->extended_value = (opline->extended_value == ZEND_INCLUDE_ONCE) ? ZEND_INCLUDE : ZEND_REQUIRE;
+        ret = apc_original_opcode_handlers[APC_OPCODE_HANDLER_DECODE(opline)](ZEND_OPCODE_HANDLER_ARGS_PASSTHRU);
+        opline->extended_value = (opline->extended_value == ZEND_INCLUDE) ? ZEND_INCLUDE_ONCE : ZEND_REQUIRE_ONCE;
+#else
         Z_LVAL(opline->op2.u.constant) = (Z_LVAL(opline->op2.u.constant) == ZEND_INCLUDE_ONCE) ? ZEND_INCLUDE : ZEND_REQUIRE;
         ret = apc_original_opcode_handlers[APC_OPCODE_HANDLER_DECODE(opline)](ZEND_OPCODE_HANDLER_ARGS_PASSTHRU);
         Z_LVAL(opline->op2.u.constant) = (Z_LVAL(opline->op2.u.constant) == ZEND_INCLUDE) ? ZEND_INCLUDE_ONCE : ZEND_REQUIRE_ONCE;
+#endif
     } else {
         ret = apc_original_opcode_handlers[APC_OPCODE_HANDLER_DECODE(opline)](ZEND_OPCODE_HANDLER_ARGS_PASSTHRU);
     }
Index: apc_zend.h
===================================================================
--- apc_zend.h	(revision 298155)
+++ apc_zend.h	(working copy)
@@ -128,7 +128,11 @@
         _UNUSED_CODE, /* 15             */
         _CV_CODE      /* 16 = IS_CV     */
     };
+#ifdef ZEND_ENGINE_2_4
+    return (opline->opcode * 25) + (apc_vm_decode[opline->op1_type] * 5) + apc_vm_decode[opline->op2_type];
+#else
     return (opline->opcode * 25) + (apc_vm_decode[opline->op1.op_type] * 5) + apc_vm_decode[opline->op2.op_type];
+#endif
 }
 
 # define APC_ZEND_OPLINE                    zend_op *opline = execute_data->opline;
Index: apc_string.c
===================================================================
--- apc_string.c	(revision 0)
+++ apc_string.c	(revision 0)
@@ -0,0 +1,291 @@
+/*
+  +----------------------------------------------------------------------+
+  | APC                                                                  |
+  +----------------------------------------------------------------------+
+  | Copyright (c) 2006-2009 The PHP Group                                |
+  +----------------------------------------------------------------------+
+  | This source file is subject to version 3.01 of the PHP license,      |
+  | that is bundled with this package in the file LICENSE, and is        |
+  | available through the world-wide-web at the following url:           |
+  | http://www.php.net/license/3_01.txt                                  |
+  | If you did not receive a copy of the PHP license and are unable to   |
+  | obtain it through the world-wide-web, please send a note to          |
+  | license@php.net so we can mail you a copy immediately.               |
+  +----------------------------------------------------------------------+
+  | Authors: Dmitry Stogov <dmitry@zend.com>                             |
+  +----------------------------------------------------------------------+
+
+   This software was contributed to PHP by Community Connect Inc. in 2002
+   and revised in 2005 by Yahoo! Inc. to add support for PHP 5.1.
+   Future revisions and derivatives of this source code must acknowledge
+   Community Connect Inc. as the original contributor of this module by
+   leaving this note intact in the source code.
+
+   All other licensing and usage conditions are those of the PHP Group.
+
+ */
+
+/* $Id: $ */
+
+#include "apc.h"
+#include "apc_globals.h"
+#include "apc_php.h"
+#include "apc_lock.h"
+
+#ifdef ZEND_ENGINE_2_4
+
+typedef struct _apc_interned_strings_data_t {
+    char *interned_strings_start;
+    char *interned_strings_end;
+    char *interned_strings_top;
+    apc_lck_t lock;
+    HashTable interned_strings;
+} apc_interned_strings_data_t;
+
+apc_interned_strings_data_t *apc_interned_strings_data = NULL;
+
+#define APCSG(v) (apc_interned_strings_data->v)
+
+static char *old_interned_strings_start;
+static char *old_interned_strings_end;
+static char *(*old_new_interned_string)(char *str, int len, int free_src);
+static void (*old_interned_strings_snapshot)(TSRMLS_D);
+static void (*old_interned_strings_restore)(TSRMLS_D);
+
+static char *apc_dummy_new_interned_string_for_php(char *str, int len, int free_src TSRMLS_DC)
+{
+    return str;
+}
+
+static void apc_dummy_interned_strings_snapshot_for_php(TSRMLS_D)
+{
+}
+
+static void apc_dummy_interned_strings_restore_for_php(TSRMLS_D)
+{
+}
+
+char *apc_new_interned_string(char *arKey, int nKeyLength TSRMLS_DC)
+{
+    ulong h;
+    uint nIndex;
+    Bucket *p;
+
+    if (arKey >= APCSG(interned_strings_start) && arKey < APCSG(interned_strings_end)) {
+        return arKey;
+    }
+
+    h = zend_inline_hash_func(arKey, nKeyLength);
+    nIndex = h & APCSG(interned_strings).nTableMask;
+
+    p = APCSG(interned_strings).arBuckets[nIndex];
+    while (p != NULL) {
+        if ((p->h == h) && (p->nKeyLength == nKeyLength)) {
+            if (!memcmp(p->arKey, arKey, nKeyLength)) {
+                return p->arKey;
+            }
+        }
+        p = p->pNext;
+    }
+   
+    if (APCSG(interned_strings_top) + ZEND_MM_ALIGNED_SIZE(sizeof(Bucket) + nKeyLength) >=
+        APCSG(interned_strings_end)) {
+        /* no memory */
+        return NULL;
+    }
+
+    p = (Bucket *) APCSG(interned_strings_top);
+    APCSG(interned_strings_top) += ZEND_MM_ALIGNED_SIZE(sizeof(Bucket) + nKeyLength);
+
+    p->arKey = (char*)(p+1);
+    memcpy(p->arKey, arKey, nKeyLength);
+    p->nKeyLength = nKeyLength;
+    p->h = h;
+    p->pData = &p->pDataPtr;
+    p->pDataPtr = p;
+
+    p->pNext = APCSG(interned_strings).arBuckets[nIndex];
+    p->pLast = NULL;
+    if (p->pNext) {
+        p->pNext->pLast = p;
+    }
+    APCSG(interned_strings).arBuckets[nIndex] = p;
+
+    p->pListLast = APCSG(interned_strings).pListTail;
+    APCSG(interned_strings).pListTail = p;
+    p->pListNext = NULL;
+    if (p->pListLast != NULL) {
+        p->pListLast->pListNext = p;
+    }
+    if (!APCSG(interned_strings).pListHead) {
+        APCSG(interned_strings).pListHead = p;
+    }
+
+    APCSG(interned_strings).nNumOfElements++;
+
+    return p->arKey;
+}
+
+static char *apc_new_interned_string_for_php(char *arKey, int nKeyLength, int free_src TSRMLS_DC)
+{
+    char *ret;
+
+    LOCK(APCSG(lock))
+    ret = apc_new_interned_string(arKey, nKeyLength TSRMLS_DC);
+    if (ret == NULL) {
+        UNLOCK(APCSG(lock));
+    	ret = arKey;
+    } else if (free_src) {
+        UNLOCK(APCSG(lock));
+        efree(arKey);
+    } else {
+        UNLOCK(APCSG(lock));
+    }
+    return ret;
+}
+
+int apc_interned_strings_disable(TSRMLS_D)
+{
+    CG(new_interned_string) = apc_dummy_new_interned_string_for_php;
+}
+
+int apc_interned_strings_enable(TSRMLS_D)
+{
+    CG(new_interned_string) = apc_new_interned_string_for_php;
+}
+
+static void apc_copy_interal_strings(TSRMLS_D)
+{
+    Bucket *p, *q;
+
+    p = CG(function_table)->pListHead;
+    while (p) {
+        if (p->nKeyLength) {
+            p->arKey = apc_new_interned_string(p->arKey, p->nKeyLength TSRMLS_CC);
+        }
+        p = p->pListNext;
+    }
+
+    p = CG(class_table)->pListHead;
+    while (p) {
+        zend_class_entry *ce = (zend_class_entry*)(p->pDataPtr);
+
+        if (p->nKeyLength) {
+            p->arKey = apc_new_interned_string(p->arKey, p->nKeyLength TSRMLS_CC);
+        }
+
+        q = ce->properties_info.pListHead;
+        while (q) {
+            zend_property_info *info = (zend_property_info*)(q->pData);
+
+            if (q->nKeyLength) {
+                q->arKey = apc_new_interned_string(q->arKey, q->nKeyLength TSRMLS_CC);
+            }
+
+            if (info->name) {
+                info->name = apc_new_interned_string(info->name, info->name_length+1 TSRMLS_CC);
+            }
+
+            q = q->pListNext;
+        }
+
+        q = ce->default_properties.pListHead;
+        while (q) {
+            if (q->nKeyLength) {
+                q->arKey = apc_new_interned_string(q->arKey, q->nKeyLength TSRMLS_CC);
+            }
+            q = q->pListNext;
+        }
+
+        q = ce->default_static_members.pListHead;
+        while (q) {
+            if (q->nKeyLength) {
+                q->arKey = apc_new_interned_string(q->arKey, q->nKeyLength TSRMLS_CC);
+            }
+            q = q->pListNext;
+        }
+
+        q = ce->function_table.pListHead;
+        while (q) {
+            if (q->nKeyLength) {
+                q->arKey = apc_new_interned_string(q->arKey, q->nKeyLength TSRMLS_CC);
+            }
+            q = q->pListNext;
+        }
+
+        q = ce->constants_table.pListHead;
+        while (q) {
+            if (q->nKeyLength) {
+                q->arKey = apc_new_interned_string(q->arKey, q->nKeyLength TSRMLS_CC);
+            }
+            q = q->pListNext;
+        }
+
+        p = p->pListNext;
+    }
+
+    p = EG(zend_constants)->pListHead;
+    while (p) {
+        if (p->nKeyLength) {
+            p->arKey = apc_new_interned_string(p->arKey, p->nKeyLength TSRMLS_CC);
+       }
+        p = p->pListNext;
+    }
+}
+
+void apc_interned_strings_init(TSRMLS_D)
+{
+    int count = APCG(shm_strings_buffer) / (sizeof(Bucket) + sizeof(Bucket*) * 2);
+
+    apc_interned_strings_data = (apc_interned_strings_data_t*)apc_sma_malloc(APCG(shm_strings_buffer));
+
+    CREATE_LOCK(APCSG(lock));
+
+    zend_hash_init(&APCSG(interned_strings), count, NULL, NULL, 1);
+    APCSG(interned_strings).nTableMask = APCSG(interned_strings).nTableSize - 1;
+    APCSG(interned_strings).arBuckets = (Bucket**)((char*)apc_interned_strings_data + sizeof(apc_interned_strings_data_t));
+   
+    APCSG(interned_strings_start) = (char*)APCSG(interned_strings).arBuckets + APCSG(interned_strings).nTableSize * sizeof(Bucket *);
+    APCSG(interned_strings_end)   = (char*)apc_interned_strings_data + APCG(shm_strings_buffer);
+    APCSG(interned_strings_top)   = APCSG(interned_strings_start);
+
+    old_interned_strings_start = CG(interned_strings_start);
+    old_interned_strings_end = CG(interned_strings_end);
+    old_new_interned_string = CG(new_interned_string);
+    old_interned_strings_snapshot = CG(interned_strings_snapshot);
+    old_interned_strings_restore = CG(interned_strings_restore);
+
+    CG(interned_strings_start) = APCSG(interned_strings_start);
+    CG(interned_strings_end) = APCSG(interned_strings_end);
+    CG(new_interned_string) = apc_dummy_new_interned_string_for_php;
+    CG(interned_strings_snapshot) = apc_dummy_interned_strings_snapshot_for_php;
+    CG(interned_strings_restore) = apc_dummy_interned_strings_restore_for_php;
+
+    apc_copy_interal_strings(TSRMLS_C);
+}
+
+void apc_interned_strings_shutdown(TSRMLS_D)
+{	
+	zend_hash_clean(CG(function_table));
+	zend_hash_clean(CG(class_table));
+	zend_hash_clean(EG(zend_constants));
+
+    CG(interned_strings_start) = old_interned_strings_start;
+    CG(interned_strings_end) = old_interned_strings_end;
+    CG(new_interned_string) = old_new_interned_string;
+    CG(interned_strings_snapshot) = old_interned_strings_snapshot;
+    CG(interned_strings_restore) = old_interned_strings_restore;
+
+	DESTROY_LOCK(APCSG(lock));
+}
+
+#endif
+
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * End:
+ * vim>600: expandtab sw=4 ts=4 sts=4 fdm=marker
+ * vim<600: expandtab sw=4 ts=4 sts=4
+ */
Index: apc_bin.c
===================================================================
--- apc_bin.c	(revision 298155)
+++ apc_bin.c	(working copy)
@@ -191,6 +191,7 @@
 
     /* swizzle op_array */
     for(i=0; i < op_array->last; i++) {
+#ifndef ZEND_ENGINE_2_4
         if(op_array->opcodes[i].result.op_type == IS_CONST) {
             apc_swizzle_zval(bd, ll, &op_array->opcodes[i].result.u.constant TSRMLS_CC);
         }
@@ -200,14 +201,23 @@
         if(op_array->opcodes[i].op2.op_type == IS_CONST) {
             apc_swizzle_zval(bd, ll, &op_array->opcodes[i].op2.u.constant TSRMLS_CC);
         }
+#endif
         switch (op_array->opcodes[i].opcode) {
             case ZEND_JMP:
+#ifdef ZEND_ENGINE_2_4
+                apc_swizzle_ptr(bd, ll, &op_array->opcodes[i].op1.jmp_addr);
+#else
                 apc_swizzle_ptr(bd, ll, &op_array->opcodes[i].op1.u.jmp_addr);
+#endif
             case ZEND_JMPZ:
             case ZEND_JMPNZ:
             case ZEND_JMPZ_EX:
             case ZEND_JMPNZ_EX:
+#ifdef ZEND_ENGINE_2_4
+                apc_swizzle_ptr(bd, ll, &op_array->opcodes[i].op2.jmp_addr);
+#else
                 apc_swizzle_ptr(bd, ll, &op_array->opcodes[i].op2.u.jmp_addr);
+#endif
         }
     }
     apc_swizzle_ptr(bd, ll, &op_array->opcodes);
Index: config.w32
===================================================================
--- config.w32	(revision 298155)
+++ config.w32	(working copy)
@@ -20,7 +20,7 @@
 	var apc_sources = 	'apc.c php_apc.c apc_cache.c apc_compile.c apc_debug.c ' + 
 				'apc_fcntl_win32.c apc_iterator.c apc_main.c apc_shm.c ' + 
 				'apc_sma.c apc_stack.c apc_rfc1867.c apc_zend.c apc_pool.c ' +
-				'apc_bin.c';
+				'apc_bin.c zend_string.c';
 
 	if(PHP_APC_FILEHITS != 'no')
 	{
Index: apc_string.h
===================================================================
--- apc_string.h	(revision 0)
+++ apc_string.h	(revision 0)
@@ -0,0 +1,52 @@
+/*
+  +----------------------------------------------------------------------+
+  | APC                                                                  |
+  +----------------------------------------------------------------------+
+  | Copyright (c) 2006-2009 The PHP Group                                |
+  +----------------------------------------------------------------------+
+  | This source file is subject to version 3.01 of the PHP license,      |
+  | that is bundled with this package in the file LICENSE, and is        |
+  | available through the world-wide-web at the following url:           |
+  | http://www.php.net/license/3_01.txt                                  |
+  | If you did not receive a copy of the PHP license and are unable to   |
+  | obtain it through the world-wide-web, please send a note to          |
+  | license@php.net so we can mail you a copy immediately.               |
+  +----------------------------------------------------------------------+
+  | Authors: Dmitry Stogov <dmitry@zend.com>                             |
+  +----------------------------------------------------------------------+
+
+   This software was contributed to PHP by Community Connect Inc. in 2002
+   and revised in 2005 by Yahoo! Inc. to add support for PHP 5.1.
+   Future revisions and derivatives of this source code must acknowledge
+   Community Connect Inc. as the original contributor of this module by
+   leaving this note intact in the source code.
+
+   All other licensing and usage conditions are those of the PHP Group.
+
+ */
+
+/* $Id: $ */
+
+#ifndef APC_STRING
+#define APC_STRING
+
+#include "apc.h"
+
+void apc_interned_strings_init(TSRMLS_D);
+void apc_interned_strings_shutdown(TSRMLS_D);
+
+void apc_interned_strings_disable(TSRMLS_D);
+void apc_interned_strings_enable(TSRMLS_D);
+
+char *apc_new_interned_string(char *arKey, int nKeyLength TSRMLS_DC);
+
+#endif
+
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * End:
+ * vim>600: expandtab sw=4 ts=4 sts=4 fdm=marker
+ * vim<600: expandtab sw=4 ts=4 sts=4
+ */
Index: config.m4
===================================================================
--- config.m4	(revision 298155)
+++ config.m4	(working copy)
@@ -221,7 +221,8 @@
                apc_signal.c \
                apc_pool.c \
                apc_iterator.c \
-               apc_bin.c "
+               apc_bin.c \
+               apc_string.c "
 
   PHP_CHECK_LIBRARY(rt, shm_open, [PHP_ADD_LIBRARY(rt,,APC_SHARED_LIBADD)])
   PHP_NEW_EXTENSION(apc, $apc_sources, $ext_shared,, \\$(APC_CFLAGS))
Index: apc_globals.h
===================================================================
--- apc_globals.h	(revision 298155)
+++ apc_globals.h	(working copy)
@@ -116,6 +116,9 @@
     HashTable *lazy_function_table;  /* lazy function entry table */
     zend_bool lazy_classes;          /* enable/disable lazy class loading */
     HashTable *lazy_class_table;     /* lazy class entry table */
+#ifdef ZEND_ENGINE_2_4
+    long shm_strings_buffer;
+#endif
 ZEND_END_MODULE_GLOBALS(apc)
 
 /* (the following declaration is defined in php_apc.c) */
